module VCR
  class RequestMatcherRegistry
    DEFAULT_MATCHERS = [:method, :uri]

    class Matcher < Struct.new(:callable)
      def matches?(request_1, request_2)
        callable.call(request_1, request_2)
      end
    end

    def initialize
      @registry = {}
      register_built_ins
    end

    def register(name, &block)
      if @registry.has_key?(name)
        warn "WARNING: There is already a VCR request matcher registered for #{name.inspect}. Overriding it."
      end

      @registry[name] = Matcher.new(block)
    end

    def [](matcher)
      @registry.fetch(matcher) do
        matcher.respond_to?(:call) ?
          Matcher.new(matcher) :
          raise_unregistered_matcher_error(matcher)
      end
    end

    def uri_without_params(*ignores)
      ignores = ignores.map { |i| i.to_s }

      lambda do |request_1, request_2|
        uri_1, uri_2 = [request_1, request_2].map do |r|
          URI(r.uri).tap do |uri|
            uri.query = uri.query.split('&').tap { |params|
              params.map! do |p|
                key, value = p.split('=')
                key.gsub!(/\[\]\z/, '') # handle params like tag[]=
                [key, value]
              end

              params.reject! { |p| ignores.include?(p.first) }
              params.map!    { |p| p.join('=') }
            }.join('&')
          end
        end

        uri_1 == uri_2
      end
    end
    alias uri_without_param uri_without_params

  private

    def raise_unregistered_matcher_error(name)
      raise Errors::UnregisteredMatcherError.new \
        "There is no matcher registered for #{name.inspect}. " +
        "Did you mean one of #{@registry.keys.map(&:inspect).join(', ')}?"
    end

    def register_built_ins
      register(:method)  { |r1, r2| r1.method == r2.method }
      register(:uri)     { |r1, r2| normalize_uri(r1.uri) == normalize_uri(r2.uri) }
      register(:host)    { |r1, r2| URI(r1.uri).host == URI(r2.uri).host }
      register(:path)    { |r1, r2| URI(r1.uri).path == URI(r2.uri).path }
      register(:body)    { |r1, r2| r1.body == r2.body }
      register(:headers) { |r1, r2| r1.headers == r2.headers }
    end

    def normalize_uri(uri)
      # TODO: find a better, less-hacky way to do this.
      if defined?(::WebMock)
        ::WebMock::Util::URI.normalize_uri(uri).to_s
      elsif defined?(VCR::Middleware::Faraday)
        # Faraday normalizes URIs by replacing '+' with '%20'
        uri.gsub('+', '%20')
      else
        uri
      end
    end
  end
end

